git: 78df56ccfcb4 - main - libthr: Use kern.stacktop for thread stack calculation.

From: Marcin Wojtas <mw_at_FreeBSD.org>
Date: Fri, 15 Oct 2021 08:23:17 UTC
The branch main has been updated by mw:

URL: https://cgit.FreeBSD.org/src/commit/?id=78df56ccfcb40013a3e6904bd6d39836220c3550

commit 78df56ccfcb40013a3e6904bd6d39836220c3550
Author:     Dawid Gorecki <dgr@semihalf.com>
AuthorDate: 2021-10-13 19:06:05 +0000
Commit:     Marcin Wojtas <mw@FreeBSD.org>
CommitDate: 2021-10-15 08:21:56 +0000

    libthr: Use kern.stacktop for thread stack calculation.
    
    Use the new kern.stacktop sysctl to retrieve the address of stack top
    instead of kern.usrstack. kern.usrstack does not have any knowledge
    of the stack gap, so this can cause problems with thread stacks.
    Using kern.stacktop sysctl should fix most of those problems.
    kern.usrstack is used as a fallback when kern.stacktop cannot be read.
    
    Rename usrstack variables to stacktop to reflect this change.
    
    Fixes problems with firefox and thunderbird not starting with
    stack gap enabled.
    
    PR: 239873
    Reviewed by: kib
    Obtained from: Semihalf
    Sponsored by: Stormshield
    MFC after: 1 month
    Differential Revision: https://reviews.freebsd.org/D31898
---
 lib/libthr/thread/thr_init.c    | 19 +++++++++++--------
 lib/libthr/thread/thr_private.h |  2 +-
 lib/libthr/thread/thr_stack.c   | 22 ++++++++++++----------
 3 files changed, 24 insertions(+), 19 deletions(-)

diff --git a/lib/libthr/thread/thr_init.c b/lib/libthr/thread/thr_init.c
index 82bde10a153e..f2a66c406e85 100644
--- a/lib/libthr/thread/thr_init.c
+++ b/lib/libthr/thread/thr_init.c
@@ -61,7 +61,7 @@ __FBSDID("$FreeBSD$");
 #include "libc_private.h"
 #include "thr_private.h"
 
-char		*_usrstack;
+char		*_stacktop;
 struct pthread	*_thr_initial;
 int		_libthr_debug;
 int		_thread_event_mask;
@@ -388,7 +388,7 @@ init_main_thread(struct pthread *thread)
 	 * resource limits, so this stack needs an explicitly mapped
 	 * red zone to protect the thread stack that is just beyond.
 	 */
-	if (mmap(_usrstack - _thr_stack_initial -
+	if (mmap(_stacktop - _thr_stack_initial -
 	    _thr_guard_default, _thr_guard_default, 0, MAP_ANON,
 	    -1, 0) == MAP_FAILED)
 		PANIC("Cannot allocate red zone for initial thread");
@@ -402,7 +402,7 @@ init_main_thread(struct pthread *thread)
 	 *       actually free() it; it just puts it in the free
 	 *       stack queue for later reuse.
 	 */
-	thread->attr.stackaddr_attr = _usrstack - _thr_stack_initial;
+	thread->attr.stackaddr_attr = _stacktop - _thr_stack_initial;
 	thread->attr.stacksize_attr = _thr_stack_initial;
 	thread->attr.guardsize_attr = _thr_guard_default;
 	thread->attr.flags |= THR_STACK_USER;
@@ -427,7 +427,7 @@ init_main_thread(struct pthread *thread)
 	thread->attr.prio = sched_param.sched_priority;
 
 #ifdef _PTHREAD_FORCED_UNWIND
-	thread->unwind_stackend = _usrstack;
+	thread->unwind_stackend = _stacktop;
 #endif
 
 	/* Others cleared to zero by thr_alloc() */
@@ -464,10 +464,13 @@ init_private(void)
 		__thr_malloc_init();
 		/* Find the stack top */
 		mib[0] = CTL_KERN;
-		mib[1] = KERN_USRSTACK;
-		len = sizeof (_usrstack);
-		if (sysctl(mib, 2, &_usrstack, &len, NULL, 0) == -1)
-			PANIC("Cannot get kern.usrstack from sysctl");
+		mib[1] = KERN_STACKTOP;
+		len = sizeof (_stacktop);
+		if (sysctl(mib, 2, &_stacktop, &len, NULL, 0) == -1) {
+			mib[1] = KERN_USRSTACK;
+			if (sysctl(mib, 2, &_stacktop, &len, NULL, 0) == -1)
+				PANIC("Cannot get kern.usrstack from sysctl");
+		}
 		env_bigstack = getenv("LIBPTHREAD_BIGSTACK_MAIN");
 		env_splitstack = getenv("LIBPTHREAD_SPLITSTACK_MAIN");
 		if (env_bigstack != NULL || env_splitstack == NULL) {
diff --git a/lib/libthr/thread/thr_private.h b/lib/libthr/thread/thr_private.h
index a5bbc5997d30..d6fb74bb4372 100644
--- a/lib/libthr/thread/thr_private.h
+++ b/lib/libthr/thread/thr_private.h
@@ -724,7 +724,7 @@ extern int __isthreaded;
  * Global variables for the pthread kernel.
  */
 
-extern char		*_usrstack __hidden;
+extern char		*_stacktop __hidden;
 
 /* For debugger */
 extern int		_libthr_debug;
diff --git a/lib/libthr/thread/thr_stack.c b/lib/libthr/thread/thr_stack.c
index b08bafdd9417..fe50bc76db5f 100644
--- a/lib/libthr/thread/thr_stack.c
+++ b/lib/libthr/thread/thr_stack.c
@@ -149,18 +149,20 @@ singlethread_map_stacks_exec(void)
 {
 	int mib[2];
 	struct rlimit rlim;
-	u_long usrstack;
+	u_long stacktop;
 	size_t len;
 
 	mib[0] = CTL_KERN;
-	mib[1] = KERN_USRSTACK;
-	len = sizeof(usrstack);
-	if (sysctl(mib, sizeof(mib) / sizeof(mib[0]), &usrstack, &len, NULL, 0)
-	    == -1)
-		return;
+	mib[1] = KERN_STACKTOP;
+	len = sizeof(stacktop);
+	if (sysctl(mib, nitems(mib), &stacktop, &len, NULL, 0) == -1) {
+		mib[1] = KERN_USRSTACK;
+		if (sysctl(mib, nitems(mib), &stacktop, &len, NULL, 0) == -1)
+			return;
+	}
 	if (getrlimit(RLIMIT_STACK, &rlim) == -1)
 		return;
-	mprotect((void *)(uintptr_t)(usrstack - rlim.rlim_cur),
+	mprotect((void *)(uintptr_t)(stacktop - rlim.rlim_cur),
 	    rlim.rlim_cur, _rtld_get_stack_prot());
 }
 
@@ -213,7 +215,7 @@ _thr_stack_alloc(struct pthread_attr *attr)
 
 	/*
 	 * Use the garbage collector lock for synchronization of the
 	 * spare stack lists and allocations from usrstack.
 	 */
 	THREAD_LIST_WRLOCK(curthread);
 	/*
@@ -249,11 +251,11 @@ _thr_stack_alloc(struct pthread_attr *attr)
 	}
 	else {
 		/*
 		 * Allocate a stack from or below usrstack, depending
 		 * on the LIBPTHREAD_BIGSTACK_MAIN env variable.
 		 */
 		if (last_stack == NULL)
-			last_stack = _usrstack - _thr_stack_initial -
+			last_stack = _stacktop - _thr_stack_initial -
 			    _thr_guard_default;
 
 		/* Allocate a new stack. */
@@ -213,7 +215,7 @@ _thr_stack_alloc(struct pthread_attr *attr)
 	/*
 	 * Use the garbage collector lock for synchronization of the
-	 * spare stack lists and allocations from usrstack.
+	 * spare stack lists and allocations from stacktop.
 	 */
 	THREAD_LIST_WRLOCK(curthread);
@@ -249,11 +251,11 @@ _thr_stack_alloc(struct pthread_attr *attr)
 	else {
 		/*
-		 * Allocate a stack from or below usrstack, depending
+		 * Allocate a stack from or below stacktop, depending
 		 * on the LIBPTHREAD_BIGSTACK_MAIN env variable.
 		 */